//
// (c)2019 Sassafras Software Inc.
//

(function () {

	var settings = {
			url: localStorage.getItem("url"),
			username: localStorage.getItem("username"),
			password: localStorage.getItem("password"),
			appid: localStorage.getItem("appid"),
			extid_computer: localStorage.getItem("extid-computer"),
			extid_product: localStorage.getItem("extid-product"),
			section: localStorage.getItem("section"),
			unique_field: localStorage.getItem("unique-field"),
			match_serial: localStorage.getItem("match-serial"),
			match_assetid: localStorage.getItem("match-assetid"),
			match_name: localStorage.getItem("match-name"),
			buffalo: localStorage.getItem("buffalo"),
			dormant: localStorage.getItem("dormant"),
			model: localStorage.getItem("model"),
			lifestage: localStorage.getItem("lifestage"),
			audits: localStorage.getItem("audits"),
			sandbox: localStorage.getItem("sandbox"),
			no_create_assets: localStorage.getItem("no-create-assets")
		};

	if (!app.scheduled && !app.interactive) {
		var last = localStorage.getItem("lastrun") || 0,
			frequency = parseInt(localStorage.getItem("frequency")) || 1,
			now = Math.floor(Date.now() / 1000);
		if (last + (frequency * 86000) < now) {
			localStorage.setItem("lastrun", now);
			var dun = process(settings);
			if (dun)
				console.error(dun);
		}
		return;
	}

	var sections = [ { label:"All Sections", value:0, }, "-" ],
		ksql = new KSQuery();
	ksql.command("select maid 92 idntname");
	ksql.execute(function (row) {
		sections.push({ value:row[0], label:row[1] });
	});
	if (sections.length == 2)
		sections.pop();

	return({
		version: 1,
		setup: {
			title: "Sync with TeamDynamix",
			id: "com.sassafras.teamdynamix",
			fields: [
				{
					id: "url",
					label: "Instance URL:",
					type: "text",
					default: settings.url,
					required: true
				},
				{
					id: "username",
					label: "Admin:",
					type: "text",
					default: settings.username,
					required: true
				},
				{
					id: "password",
					label: "Password:",
					type: "pass",
					default: settings.password,
					required: true
				},
				{
					id: "appid",
					label: "AppID:",
					type: "long",
					default: settings.appid,
					required: true
				},
				{
					id: "extid_computer",
					label: "Computer Viewer ID:",
					type: "long",
					default: settings.extid_computer
				},
				{
					id: "extid_product",
					label: "Product Viewer ID:",
					type: "long",
					default: settings.extid_product
				},
				{
					id: "section",
					label: "Process Computers in:",
					type: "menu",
					options: sections,
					default: settings.section || 0
				},
				{
					id: "unique_field",
					label: "Unique Value in TDX:",
					type: "menu",
					options: [
						{ label: "Use Best Match", value: 0 },
						"-",
						{ label: "Computer Name", value: "name" },
						{ label: "Serial Number", value: "oemserial" },
						{ label: "Tag (Asset ID)", value: "assetid" },
						"-",
						{ label: "External ID Only", value: "externalid" }
					],
					default: settings.unique_field || 0
				},
				{
					id: "match_serial",
					label: "Match by Serial Number",
					type: "bool",
					default: settings.match_serial
				},
				{
					id: "match_assetid",
					label: "Match by Asset ID",
					type: "bool",
					default: settings.match_assetid
				},
				{
					id: "match_name",
					label: "Match by Name",
					type: "bool",
					default: settings.match_name
				},
				{
					id: "no_create_assets",
					label: "Don't Create New Assets in TD",
					type: "bool",
					default: settings.no_create_assets
				},
				{
					id: "dormant",
					label: "Include Dormant Computers",
					type: "bool",
					default: settings.dormant
				},
				{
					id: "model",
					label: "Set Computer Model from KeyServer",
					type: "bool",
					default: settings.model
				},
				{
					id: "lifestage",
					label: "Set Lifecycle Stage from KeyServer",
					type: "bool",
					default: settings.lifestage
				},
				{
					id: "purchasedate",
					label: "Set Purchase Date from KeyServer",
					type: "bool",
					default: settings.purchasedate
				},
				{
					id: "warrantydate",
					label: "Set Warranty Date from KeyServer",
					type: "bool",
					default: settings.warrantydate
				},
				{
					id: "buffalo",
					label: "Use Deployed as default Status",
					type: "bool",
					default: settings.buffalo
				},
				{
					id: "audits",
					label: "Include Software Audit Relationships",
					type: "bool",
					default: settings.audits
				},
				{
					id: "sandbox",
					label: "Sync with Sandbox Instance",
					type: "bool",
					default: settings.sandbox
				},
				{
					id: "full",
					label: "Full Product Sync",
					type: "bool"
				},
				{
					id: "verbose",
					label: "Write to Log",
					type: "bool"
				}
			]
		},
		proceed: function (params) {
			localStorage.setItem("url", params.url);
			localStorage.setItem("username", params.username);
			localStorage.setItem("password", params.password);
			localStorage.setItem("appid", params.appid);
			localStorage.setItem("extid-computer", params.extid_computer);
			localStorage.setItem("extid-product", params.extid_product);
			localStorage.setItem("section", params.section);
			localStorage.setItem("unique-field", params.unique_field);
			localStorage.setItem("match-serial", params.match_serial);
			localStorage.setItem("match-assetid", params.match_assetid);
			localStorage.setItem("match-name", params.match_name);
			localStorage.setItem("buffalo", params.buffalo);
			localStorage.setItem("dormant", params.dormant);
			localStorage.setItem("model", params.model);
			localStorage.setItem("lifestage", params.lifestage);
			localStorage.setItem("audits", params.audits);
			localStorage.setItem("sandbox", params.sandbox);
			localStorage.setItem("no-create-assets", params.no_create_assets);
			var dun = process(params);
			if (dun)
				alert(dun);
		}
	});

	function process (settings) {
		bundle.load("td.js");
		var makers = bundle.load("makers.js"),
			uidpair = {
				oemserial: "SerialNumber",
				assetid: "Tag",
				name: "Name",
				externalid: "ExternalID"
			};

		app.progress.open(settings.audits ? 8 : 5);

		var url = settings.url,
			td = new TD(url, settings.appid || 30, { sandbox:settings.sandbox, verbose:!!settings.verbose }),
			creds = { username:settings.username || '', password:settings.password },
			auth = localStorage.getItem("auth"),
			dig = digest(url + "." + creds.username + "." + creds.password);
		if (auth && auth.token && auth.digest == dig)
			td.auth(auth);
		auth = td.auth();
		if (!auth)
			td.auth(creds);
		auth = td.auth();
		if (!auth) {
			td.log({ failure:"failed to get auth token for " + creds.username + " from " + url });
			return("incorrect URL or invalid credentials");
		}

		auth.digest = dig;
		localStorage.setItem("auth", auth);

		app.progress.advance();

		var standard = bundle.load("mappings.json"),
			mappings = {};
		/* remove any conditional mappings */
		for (var key in standard) {
			var mapset = localStorage.getItem(key) || standard[key];
			mappings[key] = mapset;
			for (var j = mapset.length; --j >= 0; ) {
				if (!test_conditions(mapset[j], settings))
					mapset.splice(j, 1);
			}
		}

		/* set up match fields */
		var keys = [];
		if (settings.unique_field && uidpair[settings.unique_field]) {
			keys.push(uidpair[settings.unique_field]);
		} else {
			keys.push("ExternalID");
			if (settings.match_assetid)
				keys.push("Tag");
			if (settings.match_serial)
				keys.push("SerialNumber");
			if (settings.match_name)
				keys.push("Name");
		}

		/*
			Computers (Assets)
		 */
		var view = (app && app.selection),
			filter = localStorage.getItem("filter") || "1",
			exclude = objectify(localStorage.getItem("exclude") || []),
			badasset = objectify(localStorage.getItem("badasset") || []);
		if (!view || !view.length || view.table.toLowerCase() != "computer") {
			view = new KSView("computer");
			view.select("(laud)&&(alwd>" + (settings.dormant ? "0" : "1") + ")&&(" + filter + ")" + (settings.section ? "&&(sect=" + settings.section + ")" : ""));
		}

		var assets = td.assets(keys).load(),
			fields = buildfields(mappings.assets),
			extid = settings.extid_computer || 0;
		if (extid)
			fields.push("ExternalSourceID");
		assets.setfields(fields);

		app.progress.advance();

		if (!settings.unique_field || !uidpair[settings.unique_field]) {
			assets.forEach(function (item) {
				if (item.ExternalID) {
					var obj = view.find(item.ExternalID);
					if (obj) {
						copyfields(obj, item, "KS", mappings.assets, td, makers);
						obj.commit();
					}
				}
			});
			app.progress.advance();

			var lookup = td.lookup("Status").load();
			assets.must("StatusID", lookup.matchid("In Use") || lookup.matchid("Inventory") || lookup.matchid("*"));

			view.forEach(function (obj) {
				if (obj.division && exclude[obj.division.toLowerCase()])
					return;

				var fields = { ExternalID:obj.id },
					match;
				if (settings.match_serial) {
					match = (obj.oemserial || '').trim();
					if (match && !badasset[match.toLowerCase()])
						fields.SerialNumber = match;
				}
				if (settings.match_assetid) {
					match = (obj.assetid || '').trim();
					if (match && !badasset[match.toLowerCase()])
						fields.Tag = match;
				}
				if (settings.match_name) {
					match = (obj.name || '').trim();
					if (match && !badasset[match.toLowerCase()])
						fields.Name = match;
				}

				var item = assets.find(fields) || {},
					found = item.ID,
					update = found && (item.ExternalID != obj.id);

				copyfields(item, obj, "TD", mappings.assets, td, makers);
				item.ExternalID = obj.id;
				if (extid)
					item.ExternalSourceID = extid;
				item.attributes = obj;
				if (!settings.no_create_assets || found)
					assets.add(item, update);
			});
			app.progress.advance();
			assets.commit();
		} else if (uidpair[settings.unique_field] == "ExternalID") {
			assets.forEach(function (item) {
				if (item.ExternalID) {
					var obj = view.find(item.ExternalID);
					if (obj) {
						copyfields(obj, item, "KS", mappings.assets, td, makers);
						obj.commit();
					}
				}
			});
			app.progress.advance();

			var lookup = td.lookup("Status").load();
			assets.must("StatusID", lookup.matchid("In Use") || lookup.matchid("Inventory") || lookup.matchid("*"));

			view.forEach(function (obj) {
				if (obj.division && exclude[obj.division.toLowerCase()])
					return;

				var fields = { ExternalID:obj.id },
					item = assets.find(fields) || {},
					found = item.ID,
					update = found && (item.ExternalID != obj.id);

				copyfields(item, obj, "TD", mappings.assets, td, makers);
				item.ExternalID = obj.id;
				if (extid)
					item.ExternalSourceID = extid;
				item.attributes = obj;
				if (!settings.no_create_assets || found)
					assets.add(item, update);
			});
			app.progress.advance();
			assets.commit();
		} else {
			/* first get the "best" choice for the existing unique ID values */
			var candidates = { },
				existing = { };
			view.forEach(function (obj) {
				if (obj.division && exclude[obj.division.toLowerCase()])
					return;

				var uid = (obj[settings.unique_field] || '').trim().toLowerCase();
				if (!uid || badasset[uid])
					return;

				var best = candidates[uid];
				if (!best || (new Date(best.lastlogin)).getTime() < (new Date(obj.lastlogin)).getTime())
					candidates[uid] = obj;
			});

			/* gather TD records according to ExternalID (KS ID), remove External ID if there is a duplicate */
			assets.forEach(function (item) {
				if (item.ExternalID) {
					if (!existing[item.ExternalID]) {
						existing[item.ExternalID] = item;
					} else {
						item.ExternalID = null;
						item.ExternalSourceID = 0;
						assets.add(item, true);
					}
				}
			});

			/* update TD records with the best KS record match */
			assets.forEach(function (item) {
				var uid = (item[uidpair[settings.unique_field]] || '').trim().toLowerCase();
				if (!uid)
					return;

				var obj = candidates[uid];
				if (obj) {
					copyfields(obj, item, "KS", mappings.assets, td, makers);
					obj.commit();

					var remapped = (item.ExternalID != obj.id);
					if (remapped) {
						if (existing[item.ExternalID] == item)
							existing[item.ExternalID] = null;
						/* check for ExternalID collision, remove ExternalID from the "other" record */
						var other = existing[obj.id];
						if (other && other != item) {
							other.ExternalID = null;
							other.ExternalSourceID = 0;
							assets.add(other, true);
						}
					}
					copyfields(item, obj, "TD", mappings.assets, td, makers);
					item.ExternalID = obj.id;
					if (extid)
						item.ExternalSourceID = extid;
					item.attributes = obj;
					assets.add(item, remapped);
					existing[item.ExternalID] = item;

					candidates[uid] = null;
				}

			});
			app.progress.advance();

			if (!settings.no_create_assets) {
				var lookup = td.lookup("Status").load();
				assets.must("StatusID", lookup.matchid("In Use") || lookup.matchid("Inventory") || lookup.matchid("*"));

				/* create new TD records from any KS records that had no existing match */
				for (var uid in candidates) {
					var obj = candidates[uid],
						field = uidpair[settings.unique_field];
					if (!obj || !field)
						continue;

					/* check for ExternalID collision, remove ExternalID from the "other" record */
					var other = existing[obj.id];
					if (other) {
						other.ExternalID = null;
						other.ExternalSourceID = 0;
						assets.add(other, true);
					}

					var item = copyfields({}, obj, "TD", mappings.assets, td, makers);
					item.ExternalID = obj.id;
					if (extid)
						item.ExternalSourceID = extid;
					item.attributes = obj;
					assets.add(item, true);
				}
			}

			app.progress.advance();
			assets.commit();
		}

		app.progress.advance();

		if (!settings.audits)
			return;

		/*
			reload the computers to (re-)get any new ones that were
			created so we know what the ConfigurationItemID is.
		 */
		assets = td.assets(keys).load();

		/*
			Products (Configuration Items)
		 */
		var start = Math.floor(Date.now() / 1000),
			last = localStorage.getItem("last-product") || 0;
		if (settings.full)
			last = 0;

		view = (app && app.selection);
		if (!view || !view.length || view.table.toLowerCase() != "product") {
			view = new KSView("product");
			view.select("(actv=2)&&(lmod>" + last + ")");
		}

		extid = settings.extid_product || 0;

		var cmdb = td.citems().load();
		app.progress.advance();

		view.forEach(function (obj) {
			var item = cmdb.find({
					ExternalID:obj.id,
					Name:obj.name
				}) || {};

			copyfields(item, obj, "TD", mappings.cmdb, td, makers);
			item.ExternalID = obj.id;
			if (extid)
				item.ExternalSourceID = extid;
			item.attributes = obj;
			cmdb.add(item);
		});
		cmdb.commit();
		app.progress.advance();

		localStorage.setItem("last-product", start);

		/*
			Audit (Relationships)
		 */
		var reltypeid = td.lookup("Configuration Item Relationship Types").load().getid("Installed on", {
					IsActive:true,
					Description:"Installed on",
					InverseDescription:"Has Installed"
				});

		if (reltypeid) {
			cmdb.forEach(function (prod) {
				if (prod.ExternalID && prod.ID) {
					var ksql = new KSQuery(), audit = [], gather = [];

					// get a list af all computers with this product, TD will need the CI ID
					ksql.command("select aprd 1 maid (prod='" + prod.ExternalID + "')");
					ksql.execute(function (row) {
						var computer = assets.find('' + row);
						if (computer && computer.ConfigurationItemID)
							gather.push(computer.ConfigurationItemID);
					});

					// remove duplicates
					if (gather.length > 0) {
						gather.sort(function (a, b) { return(a - b); });
						audit = [ gather[0] ];
						for (var i = 1; i < gather.length; i++) {
							if (gather[i] != gather[i - 1])
								audit.push(gather[i]);
						}
					}

					// get the current relationships for this product so we can do incremental update if possible
					var result = td.ajax({ endpoint: td.appid + "/cmdb/" + prod.ID + "/relationships" }).value || [],
						tdaud = {}, adds = [], deletes = [], entire = true, sample = null;
					for (var i = 0; i < result.length; i++) {
						var rel = result[i];
						if (rel.RelationshipTypeID == reltypeid) {
							tdaud[rel.ChildID] = rel.ID;
							sample = rel;
						}
					}
					if (!sample) {
						// nothing on TD, so just add it all
						adds = audit;
						console.debug(prod.Name + ": total " + audit.length);
					} else if (audit.length == 0) {
						// nothing on KS, so delete everything on TD with just two requests
						if (sample) {
							adds = [ sample.ChildID ];
							deletes = [ sample.ID ];
						}
						console.debug(prod.Name + ": deleting all");
					} else {
						// gather into 'adds' all items in the KS audit but not in the TD audit
						// remove from 'tdaud' anything in both...
						for (var i = 0; i < audit.length; i++) {
							var aud = audit[i];
							if (tdaud[aud])
								delete tdaud[aud];
							else
								adds.push(aud);
						}
						// ...so what remains in tdaud are items that are no longer in the audit
						for (var key in tdaud)
							deletes.push(tdaud[key]);

						console.debug(prod.Name + ": total " + audit.length + ", add " + adds.length + ", delete " + deletes.length);

						// figure out the method that will take the fewest number of API calls
						entire = (audit.length < adds.length + deletes.length);
						if (entire) {
							deletes = [];
							adds = audit;
						}
					}

					for (var i = 0; i < adds.length; i++) {
						td.ajax({
							endpoint:td.appid + "/cmdb/" + prod.ID + "/relationships?typeId=" + reltypeid + "&otherItemId=" + adds[i] + "&isParent=true&removeExisting=" + (entire ? "true" : "false"),
							method:"PUT"
						});
						entire = false;
					}

					for (var i = 0; i < deletes.length; i++) {
						td.ajax({
							endpoint:td.appid + "/cmdb/" + prod.ID + "/relationships/" + deletes[i],
							method:"DELETE"
						});
					}

				}
			});
		}

		app.progress.advance();
		app.progress.close();
	}

	function test_condition (condition, settings) {
		if (condition.charAt(0) == '!') {
			condition = condition.substr(1);
			if (settings.hasOwnProperty(condition) && settings[condition])
				return(false);
		} else {
			if (settings.hasOwnProperty(condition) && !settings[condition])
				return(false);
		}
		return(true);
	}

	function test_conditions (map, settings) {
		var conditions = map.condition;
		if (!conditions)
			return(true);

		if (typeof conditions === "string")
			conditions = conditions.split('&');
		if (Array.isArray(conditions)) {
			for (var i = 0; i < conditions.length; i++) {
				if (!test_condition(conditions[i], settings))
					return(false);
			}
		}
		return(true);
	}

	function objectify (arr) {
		var obj = {};
		for (var i = 0; i < arr.length; i++)
			obj[arr[i].toLowerCase()] = 1;
		return(obj);
	}

	function buildfields (mappings) {
		var fields = [];
		for (var i = 0; i < mappings.length; i++) {
			var map = mappings[i];
			if (!map.create) {
				if (map.dir && map.dir.indexOf("TD") < 0)
					continue;
			}
			fields.push(map.TD);
		}
		return(fields);
	}

	function copyfields (dst, src, to, mappings, td, makers) {
		var from = { KS:"TD", TD:"KS" }[to];
		if (!from)
			return(dst);

		for (var i = 0; i < mappings.length; i++) {
			var map = mappings[i],
				ffs = map[from],
				value;
			if (!map.create || dst[map[to]]) {
				if (map.dir && map.dir.indexOf(to) < 0)
					continue;
			}
			if (map.constant) {
				value = ffs;
			} else if (typeof ffs === "string") {
				value = src[ffs];
			} else {
				value = [];
				for (var j = 0; j < ffs.length; j++) {
					var fv = src[ffs[j]];
					if (map.alias && map.alias.hasOwnProperty(fv))
						fv = map.alias[fv];
					if (fv)
						value.push(fv);
				}
				value = value.join(map.sep || ' ');
			}
			if (value || map.clear) {
				if (map.alias && map.alias.hasOwnProperty(value))
					value = map.alias[value];
				if (map.lookup) {
					var lookup = td.lookup(map.lookup);
					if (lookup) {
						var props = { IsActive:true };
						if (makers[map.maker])
							makers[map.maker](props, src, dst, td);
						value = lookup.load().getid(value, props);
					}
				}
				if (value || map.clear)
					dst[map[to]] = value;
			}
		}
		return(dst);
	}

})();
